{ GIF decoder routine }

{ DECGIF.BAS by Rich Geldreich March 1992              }
{ DECGIF.PAS translated to Pascal, Paul Hurst 1995     }
{ GIFLOAD.PAS some speedups, enhancements, unit header }
{             by Charlie of Inquisition, 1998-2002     }

{ Original version of DECGIF.PAS was published in SWAG. }
{ Free Pascal Version - Platform independent       }
{ Public domain, but credits please if you use it. }

{ Last modification: 2002.11.20 }
{ For other sources or updates check out: }
{ http://scenergy.dfmk.hu/inqcoders/ }

{$MODE FPC}
Unit GIFLoad;

Interface

Const BufferLength  = 32767;
      ErrorInHeader = 1;
      ErrorInData   = 2;

Type GIFReadProc   = Procedure(Var LoadBuffer; SizeToLoad : Word; Var LoadedSize : Word);
     GIFWriteProc  = Procedure(DecodedValue : Integer);

     TGIFScreenDescriptor = Packed Record
             FileID       : Array[0..2] Of Char;
             GIFVersion   : Array[0..2] Of Char;
             Width        : Word;
             Height       : Word;
             Info         : Byte;
             BackGround   : Byte;
             EndOfSD      : Byte;
     End;
     TGIFImageDescriptor = Packed Record
             Divider     : Byte;
             ImageSide   : Word;
             ImageTop    : Word;
             ImageWidth  : Word;
             ImageHeight : Word;
             Info        : Byte;
     End;

     TPaletteItem = Packed Record
            Red   : Byte;
            Green : Byte;
            Blue  : Byte;
     End;
     TGIFPalette  = Array[0..255] Of TPaletteItem;

     PGIFDecoder   = ^TGIFDecoder;
     TGIFDecoder   = Object
        ErrorCode   : Byte;
        GetData     : GIFReadProc;
        PutData     : GIFWriteProc;
        Palette     : TGIFPalette;
        ScreenDesc  : TGIFScreenDescriptor;
        ImageDesc   : TGIFImageDescriptor;
        Constructor Init;
        Destructor  Done;
        Procedure   InitHeader;
        Procedure   Decode;
       Private
         ByteBuffer : Byte;
         Powers     : Array[0..8] Of Integer;
         Powers2    : Array[0..16] Of Integer;
         Prefix     : Array[0..4096] Of Integer;
         Suffix     : Array[0..4096] Of Integer;
         OutCode    : Array[0..1024] Of Integer;
         MaxCodes   : Array[0..12] Of Integer;
         CodeMask   : Array[0..8] Of Integer;
         Buffer     : Array[1..BufferLength] Of Byte;
         BitsPixel  : Integer;
         Function Exponent(Base, Ras : LongInt) : LongInt;
      End;

Implementation

Constructor TGIFDecoder.Init;
Var Counter      : Byte;
    Base         : Word;
    BufferValue  : Word;
Begin
  For Counter:=0 To 7 Do Powers[Counter+1]:=Exponent(2, Counter);
  Base:=4;
  For Counter:=0 To 11 Do Begin
    MaxCodes[Counter]:=Base;
    Base:=Base*2;
   End;
  Base:=1; BufferValue:=2;
  For Counter := 1 To 8 Do Begin
    CodeMask[Counter]:=Base;
    Base:=Base+BufferValue; BufferValue:=BufferValue*2;
   End;
  Base:=1;
  For Counter:=0 To 14 Do Begin
   Powers2[Counter]:=Base; Base:=Base*2;
  End;
 ErrorCode:=0;
End;

Destructor TGIFDecoder.Done;
Begin
End;

Procedure TGIFDecoder.InitHeader;
Var ReadedSize   : Word;
Begin
 FillChar(ScreenDesc,SizeOf(ScreenDesc),#0);
 GetData(ScreenDesc,SizeOf(ScreenDesc),ReadedSize);
 If SizeOf(ScreenDesc)>ReadedSize Then Begin ErrorCode:=1; Exit; End;
 If (ScreenDesc.FileID<>'GIF') Then Begin ErrorCode:=2; Exit; End;
 If (ScreenDesc.GIFVersion<>'87a') And (ScreenDesc.GIFVersion<>'89a') Then Begin ErrorCode:=3; Exit; End;
 BitsPixel:=(ScreenDesc.Info And 7)+1;
 GetData(Palette,(Exponent(2,(ScreenDesc.Info And 7)+1))*SizeOf(TPaletteItem),ReadedSize);
 If (Exponent(2,(ScreenDesc.Info And 7)+1))*SizeOf(TPaletteItem)>ReadedSize Then Begin ErrorCode:=4; Exit; End;
 FillChar(ImageDesc,SizeOf(ImageDesc),#0);
 GetData(ImageDesc,SizeOf(ImageDesc),ReadedSize);
 If SizeOf(ImageDesc)>ReadedSize Then Begin ErrorCode:=5; Exit; End;
 If ImageDesc.Divider<>Ord(',') Then Begin ErrorCode:=6; Exit; End;
 If ((ImageDesc.Info And 128) = 128) Or ((ImageDesc.Info And 64) = 64) Then Begin ErrorCode:=7; Exit; End;
End;

Procedure TGIFDecoder.Decode;
Var Counter    : Integer;
    ReadedSize : Word;
    CodeSize, ClearCode,EOFCode, FirstFree, InitCodeSize,MaxCode,BitMask,
    FreeCode, BlockLength, BitsIn, BytesLeft, Num, OutCount, Code,
    Address, TempChar, CurCode, OldCode, FinChar, InCode : Integer;
Begin
 GetData(ByteBuffer,SizeOf(ByteBuffer),ReadedSize);
 CodeSize := ByteBuffer;
 ClearCode := Powers2[CodeSize];
 EOFCode := ClearCode + 1; FirstFree := ClearCode + 2;
 FreeCode := FirstFree; CodeSize := CodeSize + 1;
 InitCodeSize := CodeSize; MaxCode := MaxCodes[CodeSize - 2];
 BitMask := CodeMask[BitsPixel];
 GetData(ByteBuffer,SizeOf(ByteBuffer),ReadedSize);
 BlockLength := ByteBuffer + 1;
 BitsIn := 8; BytesLeft := 0; Num := 0;
 OutCount := 0;
 Repeat
   Code:=0;
   For Counter:=0 To CodeSize-1 Do Begin
     Inc(BitsIn);
     If BitsIn=9 Then Begin
       Dec(BytesLeft);
       If BytesLeft<=0 Then Begin
         GetData(Buffer,SizeOf(Buffer),ReadedSize);
         BytesLeft:=ReadedSize;
         Address:=0;
       End;
       Inc(Address);
       TempChar:=Buffer[Address];
       BitsIn:=1;
       Inc(Num);
       If Num=BlockLength Then Begin
         Dec(BytesLeft);
         If BytesLeft<=0 Then Begin
           GetData(Buffer,SizeOf(Buffer),ReadedSize);
           BytesLeft := ReadedSize;
           Address := 0;
         End;
         BlockLength:=TempChar+1;
         Inc(Address);
         TempChar:=Buffer[Address];
         Num:=1;
       End;
     End;
     If (TempChar And Powers[BitsIn])>0 Then Code:=Code+Powers2[Counter];
   End;

   If Code<>EOFCode Then Begin
     If Code=ClearCode Then Begin
       CodeSize:=InitCodeSize;
       MaxCode:=MaxCodes[CodeSize-2];
       FreeCode:=FirstFree;
       Code:=0;
       For Counter:=0 To CodeSize-1 Do Begin
         Inc(BitsIn);
         If BitsIn=9 Then Begin
           Dec(BytesLeft);
           If BytesLeft<=0 Then Begin
             GetData(Buffer,SizeOf(Buffer),ReadedSize);
             Address:=0;
             BytesLeft:=ReadedSize;
           End;
           Inc(Address);
           TempChar:=Buffer[Address];
           BitsIn:=1;
           Inc(Num);
           If Num=BlockLength Then Begin
             Dec(BytesLeft);
             If BytesLeft<=0 Then Begin
               GetData(Buffer, SizeOf(Buffer),ReadedSize);
               Address:=0;
               BytesLeft:=ReadedSize;
             End;
             BlockLength:=TempChar+1;
             Inc(Address);
             TempChar:=Buffer[Address];
             Num:=1;
           End;
         End;
         If (TempChar And Powers[BitsIn])>0 Then Begin Code:=Code+Powers2[Counter];
       End;
     End;
     CurCode := Code;
     OldCode := Code;
     FinChar := Code And BitMask;
     PutData(FinChar);
   End
   Else Begin
     CurCode:=Code;
     InCode:=Code;
     If Code>=FreeCode Then Begin
       CurCode:=OldCode;
       OutCode[OutCount]:=FinChar;
       Inc(OutCount);
     End;
     If CurCode>BitMask Then Begin
       Repeat
         OutCode[OutCount]:=Suffix[CurCode];
         Inc(OutCount);
         CurCode:=PreFix[CurCode];
       Until CurCode<=BitMask;
     End;
     FinChar:=CurCode And BitMask;
     OutCode[OutCount]:=FinChar;
     Inc(OutCount);
     For Counter:=OutCount-1 DownTo 0 Do Begin
       PutData(OutCode[Counter]);
     End;
     OutCount:=0;
     PreFix[FreeCode]:=OldCode;
     Suffix[FreeCode]:=FinChar;
     OldCode:=InCode;
     Inc(FreeCode);
     If (FreeCode>=MaxCode) And (CodeSize<12) Then Begin
       Inc(CodeSize);
       Inc(MaxCode,MaxCode);
     End;
   End;
  End;
 Until Code = EOFCode;
End;

Function TGIFDecoder.Exponent(Base, Ras : LongInt) : LongInt;
Var Counter, ValueBuffer : LongInt;
Begin
 If Ras = 0 Then Begin Exponent:=1; Exit; End;
 If Ras = 1 Then Begin Exponent:=Base; Exit; End;
 ValueBuffer:=Base;
 For Counter:=1 To Ras-1 Do ValueBuffer:=ValueBuffer * Base;
 Exponent:=ValueBuffer;
End;

Begin
End.

